package body ITI8590.Sorting is

   procedure Swap(Data : in out List; Index_1, Index_2 : in Index) is
      Memo : Element;
   begin
      Memo := Data(Index_1);
      Data(Index_1) := Data(Index_2);
      Data(Index_2) := Memo;
   end Swap;

   procedure Selection_Sort(Data : in out List) is
   begin
      for Destination_Index in Data'Range loop
         for Source_Index in Destination_Index + 1 .. Data'Last loop
            if Data(Source_Index) < Data(Destination_Index) then
               Swap(Data, Source_Index, Destination_Index);
            end if;
         end loop;
      end loop;
   end Selection_Sort;

   procedure Merge(Source_1, Source_2 : in List; Destination : out List) is
      Source_1_Index    : Index := Source_1'First;
      Source_2_Index    : Index := Source_2'First;
      Destination_Index : Index := Destination'First;
   begin
      while Source_1_Index <= Source_1'Last and
            Source_2_Index <= Source_2'Last loop
         if Source_1(Source_1_Index) < Source_2(Source_2_Index) then
            Destination(Destination_Index) := Source_1(Source_1_Index);
            Source_1_Index := Source_1_Index + 1;
         else
            Destination(Destination_Index) := Source_2(Source_2_Index);
            Source_2_Index := Source_2_Index + 1;
         end if;
         Destination_Index := Destination_Index + 1;
      end loop;
      if Source_1_Index <= Source_1'Last then
         Destination(Destination_Index .. Destination'Last)
            := Source_1(Source_1_Index .. Source_1'Last);
      else
         Destination(Destination_Index .. Destination'Last)
            := Source_2(Source_2_Index .. Source_2'Last);
      end if;
   end Merge;

   procedure Merge_All(Source      : in  List;
                       Destination : out List;
                       Block_Size  : in  Index) is
      Source_1_Index    : Index := Source'First;
      Source_2_Index    : Index := Source'First + Block_Size;
      Destination_Index : Index := Destination'First;
      function Block_End(Block_Start : in Index) return Index is
      begin
         return Block_Start + Block_Size - 1;
      end Block_End;
   begin
      while Block_End(Source_2_Index) <= Source'Last loop
         Merge(Source(Source_1_Index .. Block_End(Source_1_Index)),
               Source(Source_2_Index .. Block_End(Source_2_Index)),
               Destination(Destination_Index ..
                           Destination_Index + 2 * Block_Size - 1));
         Source_1_Index    := Source_1_Index    + 2 * Block_Size;
         Source_2_Index    := Source_2_Index    + 2 * Block_Size;
         Destination_Index := Destination_Index + 2 * Block_Size;
      end loop;
      if Block_End (Source_1_Index) <= Source'Last then
         Merge(Source(Source_1_Index .. Block_End(Source_1_Index)),
               Source(Source_2_Index .. Source'Last),
               Destination(Destination_Index .. Destination'Last));
      else
         Destination(Destination_Index .. Destination'Last)
            := Source(Source_1_Index .. Source'Last);
      end if;
   end Merge_All;

   procedure Merge_Sort(Data : in out List) is
      Buffer : List(Data'Range);
      Block_Size : Index := 1;
   begin
      while Block_Size < Data'Length loop
         Merge_All(Data, Buffer, Block_Size);
         Block_Size := 2 * Block_Size;
         Merge_All(Buffer, Data, Block_Size);
         Block_Size := 2 * Block_Size;
      end loop;
   end Merge_Sort;

end ITI8590.Sorting;
